"""
Peptacular: A ProForma peptide sequence parser and annotation library
"""

from tacular import *

from .annotation import *
from .chem import *
from .constants import *
from .digestion import *
from .fasta import *
from .isotope import *
from .proforma_components import *
from .property import *
from .regex_utils import *
from .sequence import *
from .spans import *
from .utils import *

__version__ = "3.0.0"
